#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <unistd.h>
#include "hostc.h"

/* Put here the pathname of your .plan fifo */
#define FIFO      "/prism2/tcc/dragon/.plan"

/* put here the pathname of the text file to display for remote sites */
#define REMOTE    "/prism2/tcc/dragon/.bigplan"

/* put here the pathname of the text file to display for local sites */
#define LOCAL     "/prism2/tcc/dragon/.shortplan"

/* put here the pathname of the file that will contain the notification
   entries */
#define NOTIFY    "/prism2/tcc/dragon/.plannotify"

/* put here the pathname and flags of the program which displays the
   socket information .. for most unix systems, this is "netstat -n" */
#define NETSTAT   "netstat -n"

/* the port number for finger; usually ".79 " (put the '.' and extra space
   in there) */
#define FPORT     ".79 "

#define MAXLINE 256

/***********************************************************************/
/* extract_addr takes as its input a line from 'netstat -n' which is known
   to contain the string "ESTABLISHED" and an address indicating the finger
   port is being used (ending in ".79 " .. space is to enforce the *end* of
   the address).

   extract_addr returns the address which is doing the fingering, minus the
   port number from which it is fingering (irrelevant and random anyway).
   
   EXAMPLE: 
   input:
   tcp    0    0  129.138.1.5.79       129.138.1.5.3688       ESTABLISHED
   return:
   129.138.1.5
   */   
/***********************************************************************/

char *extract_addr(input)
char *input;
 {
  static char temp[MAXLINE];	/* made static because of problems I was
								   having with memory allocation */
  input=strdup(strstr(input,FPORT)); /* input points to the first
										 occurrance of '.79 ' */
  sscanf(input,"%*[.0-9]%*[ \t]%[.0-9]",temp); /* read in digits and .'s
												and whitespace following
												(no assignment) 
												assign remaining digits and 
												dots to temp */
  *(rindex(temp,'.'))='\0';		/* replace the last . in the address with
								   NULL, effectively clipping the irrelevant
								   port number*/
  return temp; }				/* return the resulting address */


/***********************************************************************/
/* printout takes as its arguments file descriptors 'plan' and 'realplan'
   which represent the named-pipe file descriptor, and the .plan text file
   descriptor, respectively.
   
   printout reads in each line from 'realplan' and sends it to the FIFO
   'plan', stopping at EOF.
   */
/***********************************************************************/

void printout(plan,realplan)
int plan;
int realplan;
{
  char line[MAXLINE];			/* used to store each line temporarily */
  FILE *frealp;					/* needed because writing lines to file
								   descriptor gave bad results
								   (garbage text) */
  frealp=fdopen(realplan,"r"); 

  while(fgets(line,MAXLINE,frealp))	/* read in a line, send it out */
	write(plan,line,strlen(line));
  fclose(frealp);
}
	

/***********************************************************************/
/* fing takes as its input a string representing the site name (in name
   format, e.g.: minos.nmt.edu) and a file descriptor 'file'.

   fing will finger the site given by 'site', and print the header lines
   to the file 'file', as well as any line showing a user who is non-idle
   (idle less than reportable by the remote finger program).

   'file' will be appended to, and (?) opened if non existent.
   */
/***********************************************************************/

void fing(site,file)
char *site;
int file;
{
  FILE *in;						/* stream used for 'finger' command */
  char tmpbuf[MAXLINE], *tmp, *tmp2; /* various kludgy strings needed
										to make this thing work right */
  int idleloc=0, count;			/* idleloc represents the location in the
								   finger-lines where the idle-time info
								   begins; count is used to count the 
								   whitespace in that field */
  tmp=(char *)malloc(40);		/* give tmp space */
  strcpy(tmp,"finger @");		/* copy 'finger @' into tmp */
  tmp2=strdup(strcat(tmp,site)); /* give tmp2 space equal to
								  that resuired by tmp+site; 
								  tmp2 becomes something like: 
								  finger @minos.nmt.edu */
  /* it would have been easier just to say:
	 tmp2=strdup(strcat("finger @",site)), but for some reason, this wouldn't
	 even compile (or at least, it caused a dump). */

  /* open stream to 'finger' command */
  if (!(in=popen(tmp2,"r")) ) {	
	fprintf(stderr,"error popening finger @%s\n",site);
	exit(0); }

  while(fgets(tmpbuf,MAXLINE,in)&&!ferror(in)) {
	count=0;					/* reset the count */
	/* ugly way to do this, but strstr is case sensitive, and old-C wouldn't
	   allow: tmp=strstr(blah) || tmp=strstr(Blah); */

	if (strstr(tmpbuf,"Idle") || strstr(tmpbuf,"idle")) {
	  if (strstr(tmpbuf,"idle"))
		/* gee, this is silly.  Works in the condition above, but not
		   in the statement below.  I needed a strdup, apparently.  Makes
		   no sense, but it works */
		tmp=strstr(tmpbuf,strdup("idle"));
	  else
		tmp=strstr(tmpbuf,strdup("Idle"));

	  idleloc=tmp-tmpbuf;		/* idle loc is the relative position of the
								   start of the 'idle' field */
	  write(file,"\t",1);		/* send a tab to the record file */
	  write(file,tmpbuf,strlen(tmpbuf)); } /* send this header-line */

	if (!idleloc) {				/* if idleloc hasn't been assigned yet,
								   we've not received the header-line yet, so
								   print out whatever's going on.  Example: 
								   [minos.nmt.edu] Connection refused. */
	  write(file,"\t",1);		/* tab it for neatness */
	  write(file,tmpbuf,strlen(tmpbuf)); } /* send the line, whatever it is */

	else {						/* idleloc is defined,
								   so examine the idle-field */
	  /* don't ask how this works, it just does.  while the characters
		 starting from the beginning of the field are non-printing,
		 increment the counter */
	  while(!isgraph(*(tmpbuf+idleloc+(count++))));
	  if (count > 4) {			/* if the count is more than 4 (size of
								   the idle field) then there's no info
								   in the field, ==> the user is non-idle */
		write(file,"\t",1);		/* so send neatness */
		write(file,tmpbuf,strlen(tmpbuf)); } } /* and the line */
  }
  /* nice separation designator between records */
  write(file,"--------------------------------------------------\n",51);
  pclose(in);					/* close the process stream! */
}


/***********************************************************************/
/*  Yay! The function we've all been waiting for.  Prepare for ugliness */
main() {
  FILE *netstat;			/* used for the 'netstat -n' process stream */
  struct hostent *hostptr;		/* stores the host information for when
								   we convert ip number to ip name */
  int plan,notify,bigplan,shortplan; /* file descriptors */
  char *inbuf, *extra;			/* for reading lines from netstat; temp */
  int i=1, found=0;				/* is 'i' ever used?  found ==> finger-port
					    		  line in netstat discovered - stop reading */
  char *time_of_finger, *tmp;	/* time stamp, and another temp string */


  inbuf=(char *)malloc(MAXLINE); /* allocate MAXLINE spaces for inbuf */

  /* yep! a reason for an infinite loop .. imagine that */
  while(1) { 

	/* these all should be put in a function call, but: */

	/* open FIFO for writing only.  program hangs here until the FIFO
	   is opened for reading (that's how they work .. neat, huh?) */
	if ((plan=open(FIFO, O_WRONLY )) == -1) { 
	  perror("open() failed");
	  exit(1); }

	/* once the FIFO is opened for reading, we know someone's trying to finger
	   here, so get rolling.  Open a descriptor to the actual .plan, soon to
	   be much more user-friendly (like, default is $HOME environment
	   variable, filename: .plantext) */
	/* note the flags passed: read-only, non-blocking (don't wait
	   for anything .. this flag might not be needed, though) */
	if ((bigplan=open(REMOTE, O_RDONLY | O_NONBLOCK)) == -1) {
	  perror("open() failed");
	  exit(1); }
	if ((shortplan=open(LOCAL, O_RDONLY | O_NONBLOCK)) == -1) {
	  perror("open() failed");
	  exit(1); }


	/* notify is the descriptor used for the 'record' file, where
	   the finger information and timestamp will be placed */
	if ((notify=open(NOTIFY, O_APPEND | O_WRONLY )) == -1) { 
	  perror("open() failed");
	  exit(1); }

	/* stream setup for the netstat -n command */
	if (!(netstat=popen(NETSTAT,"r")) ) {
	  fprintf(stderr,"error popening netstat -n\n");
	  exit(0); }
	
	/* Keep reading netstat until the appropriate string is
	   'found'.  Is this MAXLINE-1 necessary?  can't it be
	   just MAXLINE?  Probably. */
	while (fgets(inbuf,MAXLINE-1,netstat)&&!found&&!ferror(netstat)) {
	  /* if there's a finger port active, move on it */
	  if (strstr(inbuf,FPORT) && strstr(inbuf,"ESTABLISHED")) {
		/* extract the relevant address from this line */
		extra=strdup(extract_addr(inbuf));

		/* if get_host finds the address in the db */
		if (hostptr=get_host(extra)) {
		  /* AND actually gives us a name to play with */
		  if (hostptr->h_name) {
			/* send the hostname to the 'record' file */
			write(notify,hostptr->h_name,strlen(hostptr->h_name));
			write(notify,"\t",1);

			/* time stamp it right after that */
			system("echo `date` >> /prism2/tcc/dragon/.plannotify");

			/* the way that *should* have worked */
/*			if ((*tp=time()) != -1) {
			  time_of_finger=strdup(ctime(tp));
			  write(notify,strcat("\t",time_of_finger),
					strlen(time_of_finger)+1); }
			else
			  write(notify,"\ttime NA\n",9);  */

			/* finger the host, send info to 'record' file */
			fing(hostptr->h_name,notify);

			/* kludgy way of sending:
			   HELLO <hostname>!
			   Enjoy this .plan :)
			   */
			tmp=strdup("\n\tHELLO ");
			write(plan,tmp,strlen(tmp));
			/* gee, you'd think:
			   write(plan,strdup("\n\tHELLO "),8);
			   would work.  But noooooo. */

			write(plan,hostptr->h_name,strlen(hostptr->h_name));
			tmp=strdup("!\n\tEnjoy this .plan :)\n");
			write(plan,tmp,strlen(tmp));
		  }

		  /* if get_host didn't return a host, terminate.
			 okay, that's a little nasty, we should close
			 everything up, and still show them the unagumented
			 .plantext, but ... */
		  else {
			perror("get_host did not return hostname");
			exit(1); } } 

		found=1;  }
	}

	/* if we didn't find '.79' in netstat -n, then either
	   A) the finger is being done from the same machine,
	      i.e. - 'finger dragon' OR:
	   B) the file is being viewed directly */
	if (!found) {
	  write(plan,
			"You're not fingering this file, you're just reading it\n",55);
	}
	
	/* send the .plantext to the FIFO */
	if (!strstr(hostptr->h_name,"nmt.edu"))
	  printout(plan,bigplan);
	else
	  printout(plan,shortplan);
	  

	/* close up shop */
	pclose(netstat);
	close(bigplan);
	close(shortplan);
	close(plan); 
	close(notify);
	found=0;

	/* this may not be necessary, but I found that when
	   the loop was repeated, sometimes it thought the .plan
	   was still being read.  So wait a little bit for the
	   'finger' to terminate */
	sleep (5);
  }
}
